/* Copyright (c) 2022 Daniele Lombardi / Daniels118
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package it.ld.sh3d.devtoolkit;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.file.FileSystems;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;

import com.eteks.sweethome3d.plugin.Plugin;
import com.eteks.sweethome3d.plugin.PluginAction;

import static java.nio.file.StandardWatchEventKinds.ENTRY_CREATE;
import static java.nio.file.StandardWatchEventKinds.ENTRY_DELETE;
import static java.nio.file.StandardWatchEventKinds.ENTRY_MODIFY;

/* Change history
 * 2022-01-28 - v1.1   - Added support for multiple paths to monitor
 * 					   - Bug fixes
 * 2022-01-11 - v1.0   - First release.
 * 2021-12-08 - v0.1   - Alpha version.
 */

/**This class allows to load a plugin directly from .class files in a directory, and automatically reload all the files if a modification
 * occurrs in the directory. This permits to load the plugins without having to rebuild the sh3p package, and allows to see the changes
 * without having to restart SweetHome3D, enabling for quick trial and error.<br/>
 * This system can be easily integrated in existing plugins. In order to activate the dynamic class loading you have to:
 * <ol>
 *  <li>Import PluginDevToolkit.zip into source folder of your plugin project.</li>
 *  <li>
 *   Create a new class named "<i>YourPlugin</i>Loader" which extends DynamicPluginLoader and override its constructor as
 *   <pre>public <i>YourPlugin</i>Loader() {
 *    super("<i>your.package.YourPlugin</i>", new String[] {"<i>your.package</i>"});
 *}</pre>
 *  </li>
 *  <li>
 *   Edit the ApplicationPlugin.properties and change the value of the "class" parameter to the fully qualified name of the class you created
 *   in the previous step (eg. "<i>your.package.YourPlugin</i>Loader").
 *  </li>
 *  <li>Build your .sh3p package just once and place it in the SH3D plugins folder.</li>
 *  <li>
 *   Add to your run configuration a Java property named "plugin.classpath" pointing to the output folder of your class files
 *   (usually <i>your_project_folder</i>/bin). If you need to reference multiple paths you can concatenate them using the path separator.
 *  </li>
 * </ol>
 * <p>
 *  You can then run your configuration withour rebuilding the sh3p package, and even test your changes while SH3D is running.<br/>
 *  When you're ready to release your plugin to the public there is no need to remove the customizations, you can safely leave them in place,
 *  but <b>don't forget to rebuild the sh3p package before distributing it!</b>
 * </p>
 * <h3>Lifecycle management</h3>
 * <p>
 *  To avoid unexpected results when reloading the plugin at runtime, you should correctly implement the plugin lifecycle management by
 *  overriding the init() and destroy() methods in your real plugin class. Everytime a change is detected in class files, the
 *  DynamicPluginLoader calls the destroy() method to let you release any allocated resource; your plugin is then replaced by a fresh new
 *  instance of the new class and the init() method is called to let you initialize your plugin.
 * </p>
 * <h3>Limitations</h3>
 * <ul>
 *  <li>
 *   If you add or remove an Action from the values returned by getActions() method, you have to rerun your configuration before the change
 *   takes effect. Existing actions will be updated in a seamless way.
 *  </li>
 * </ul>
 * @author Daniels118
 *
 */
public abstract class DynamicPluginLoader extends Plugin {
	private static ClassLoader classLoader = null;
	
	private final String classpath;
	private final String[] basePackageNames;
	private final String pluginClassName;
	
	private WatchService watchService;
	private ScheduledExecutorService scheduler = Executors.newSingleThreadScheduledExecutor();
	private ScheduledFuture<?> updateTask;
	
	private Plugin plugin;
	private Map<String, DynamicAction> dynamicActions = null;
	
	/**
	 * @param pluginClassName the fully qualified name of the real plugin class
	 * @param basePackageNames an array of packages prefix whose classes must be reloaded in case of modifications
	 */
	public DynamicPluginLoader(String pluginClassName, String[] basePackageNames) {
		super();
		this.classpath = System.getProperty("plugin.classpath");
		this.basePackageNames = basePackageNames;
		this.pluginClassName = pluginClassName;
		loadPlugin();
		if (classpath != null) {
			try {
				watchService = FileSystems.getDefault().newWatchService();
				String[] paths = classpath.split(Pattern.quote(File.pathSeparator));
				for (String path : paths) {
					File dir = new File(path);
					registerRecursive(dir.toPath());
				}
				Executors.newSingleThreadExecutor().submit(whatcher);
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else {
			System.out.println("DynamicPluginLoader: plugin.classpath property not set");
		}
	}
	
	private void registerRecursive(final Path root) throws IOException {
	    Files.walkFileTree(root, new SimpleFileVisitor<Path>() {
	        @Override
	        public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) throws IOException {
	            dir.register(watchService, ENTRY_CREATE, ENTRY_DELETE, ENTRY_MODIFY);
	            return FileVisitResult.CONTINUE;
	        }
	    });
	}
	
	private final Runnable whatcher = new Runnable() {
		public void run() {
			System.out.println("DynamicPluginLoader: now monitoring "+classpath);
			while (true) {
	            if (Thread.interrupted()) break;
	            WatchKey key = null;
	            try {
	                key = watchService.take();
	                key.pollEvents();
	                System.out.println("DynamicPluginLoader: changes detected within "+classpath);
	            } catch (InterruptedException ie) {
	            	Thread.currentThread().interrupt();
	                break;
	            }
	            boolean valid = key.reset();
	            if (!valid) {
	                System.err.println("DynamicPluginLoader: "+classpath+" is inaccessible and will be no longer monitored.");
	                break;
	            }
	            if (updateTask != null) updateTask.cancel(false);
	            updateTask = scheduler.schedule(updater, 500, TimeUnit.MILLISECONDS);
	        }
	    }
	};
	
	private final Runnable updater = new Runnable() {
		public void run() {
			updateTask = null;
			plugin.destroy();
            reloadPlugin();
            plugin.init();
            getActions();
		}
	};
	
	private void loadPlugin() {
		if (classLoader == null) {
			ClassLoader parent = this.getClass().getClassLoader();
			if (classpath != null) {
				System.out.println("DynamicPluginLoader: loading "+pluginClassName+" from "+classpath);
				classLoader = new MyClassLoader(parent, classpath, basePackageNames);
			} else {
				classLoader = parent;
			}
		}
		try {
			Class<?> pluginClass = classLoader.loadClass(pluginClassName);
			plugin = (Plugin) pluginClass.newInstance();
			setProperties();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	private void reloadPlugin() {
		System.out.println("DynamicPluginLoader: reloading "+pluginClassName+" from "+classpath);
		ClassLoader parent = this.getClass().getClassLoader();
		classLoader = classpath != null ? new MyClassLoader(parent, classpath, basePackageNames) : parent;
		try {
			Class<?> pluginClass = classLoader.loadClass(pluginClassName);
			plugin = (Plugin) pluginClass.newInstance();
			setProperties();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	@Override
	public final void init() {
		if (plugin != null) {
			setProperties();
			plugin.init();
		} else {
			System.err.println(pluginClassName + " not loaded");
		}
	}
	
	@Override
	public final PluginAction[] getActions() {
		if (plugin != null) {
			PluginAction[] actions = plugin.getActions();
			if (dynamicActions == null) {
				dynamicActions = new HashMap<String, DynamicAction>();
				for (PluginAction action : actions) {
					String className = action.getClass().getName();
					DynamicAction dynamicAction = new DynamicAction(action);
					dynamicActions.put(className, dynamicAction);
				}
			} else {
				for (PluginAction action : actions) {
					String className = action.getClass().getName();
					DynamicAction dynamicAction = dynamicActions.get(className);
					if (dynamicAction != null) {
						dynamicAction.setAction(action);
						System.out.println("Action "+className+" reloaded");
					} else {
						System.err.println("New action "+className+" won't be added until SH3D is restarted");
					}
				}
			}
			return dynamicActions.values().toArray(new PluginAction[dynamicActions.size()]);
		} else {
			System.err.println(pluginClassName + " not loaded");
			return new PluginAction[0];
		}
	}
	
	@Override
	public final void destroy() {
		if (plugin != null) {
			plugin.destroy();
		} else {
			System.err.println(pluginClassName + " not loaded");
		}
	}
	
	private void setProperties() {
		set("pluginClassLoader", classLoader);
		set("name", this.getName());
		set("description", this.getDescription());
		set("version", this.getVersion());
		set("license", this.getLicense());
		set("provider", this.getProvider());
		set("userPreferences", this.getUserPreferences());
		set("home", this.getHome());
		set("homeController", this.getHomeController());
		set("undoableEditSupport", this.getUndoableEditSupport());
	}
	
	private void set(String name, Object value) {
		try {
			Field field = Plugin.class.getDeclaredField(name);
			field.setAccessible(true);
			field.set(plugin, value);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
