/* Copyright (c) 2022 Daniele Lombardi / Daniels118
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package it.ld.sh3d.devtoolkit;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Pattern;

/* Change history
 * 2022-01-28 - Added support for multiple paths
 * 2022-01-21 - Fixed findResources bug occurring when there was another plugin using the library and plugin.classpath property is set.
 * 2022-01-13 - First release.
 */

class MyClassLoader extends ClassLoader {
	private final String[] classpath;
	private final String[] dynamicBasePackages;
	private final String[] dynamicBaseDirs;
	
	/**Constructs a class loader that searches for classes in a given directory. Only classes whose package name matches a list of
	 * prefixes are loaded from the specified directory. The loading of other classes is delegated to a parent class loader.
	 * @param parent the parent class loader to delegate loading classes from packages not handled by this loader
	 * @param classpath the root directory from which the search for classes starts. Multiple paths can be specified using the path separator.
	 * @param dynamicBasePackages an array of package prefixes that must be handled by this loader
	 */
	public MyClassLoader(ClassLoader parent, String classpath, String[] dynamicBasePackages) {
        super(parent);
        String[] paths = classpath.split(Pattern.quote(File.pathSeparator));
        for (int i = 0; i < paths.length; i++) {
        	if (!paths[i].endsWith(File.separator)) paths[i] += File.separator;
        }
        this.classpath = paths;
        this.dynamicBasePackages = new String[dynamicBasePackages.length];
        this.dynamicBaseDirs = new String[dynamicBasePackages.length];
        for (int i = 0; i < dynamicBasePackages.length; i++) {
        	String dynamicBasePackage = dynamicBasePackages[i];
        	if (!dynamicBasePackage.endsWith(".")) dynamicBasePackage += ".";
            this.dynamicBasePackages[i] = dynamicBasePackage;
            this.dynamicBaseDirs[i] = resourceToPath(dynamicBasePackage);
        }
    }
	
	@Override
    public Class<?> loadClass(String name) throws ClassNotFoundException {
        if (!startsWith(name, dynamicBasePackages)) return this.getClass().getClassLoader().loadClass(name);
        Class<?> cls = findLoadedClass(name);
        if (cls != null) return cls;
        for (String path : classpath) {
        	File file = new File(path + nameToPath(name) + ".class");
            if (file.exists()) {
    	        try {
    	        	byte[] classData = readFile(file);
    	            return defineClass(name, classData, 0, classData.length);
    	        } catch (IOException e) {
    	        	throw new ClassNotFoundException("Class "+name+" not found", e);
    	        }
            }
        }
        return this.getClass().getClassLoader().loadClass(name);
    }
    
	@Override
	public URL getResource(String name) {
		if (!startsWith(name, dynamicBaseDirs)) return super.getResource(name);
		URL r = findResource(name);
		if (r != null) return r;
		return super.getResource(name);
	}
	
	@Override
	public Enumeration<URL> getResources(String name) throws IOException {
		if (!startsWith(name, dynamicBaseDirs)) return super.getResources(name);
		Enumeration<URL> r = findResources(name);
		if (r.hasMoreElements()) return r;
		return super.getResources(name);
	}
	
    @Override
    protected URL findResource(String name) {
    	try {
    		Enumeration<URL> resources = findResources(name);
    		if (resources.hasMoreElements()) return resources.nextElement();
    	} catch (IOException e) {
    		e.printStackTrace();
    	}
		return super.findResource(name);
    }
    
    @Override
    protected Enumeration<URL> findResources(String name) throws IOException {
    	String[] tokens = name.split("/");
    	for (String path : classpath) {
    		String packageDir = path + join(tokens, File.separator, tokens.length - 1) + File.separator;
        	final String filename = tokens[tokens.length - 1];
        	File file = new File(packageDir + filename);
        	if (file.exists()) {
        		Set<URL> urls = new HashSet<URL>();
        		urls.add(file.toURI().toURL());
        		return Collections.enumeration(urls);
        	}
    	}
    	return super.findResources(name);
    }
    
    private byte[] readFile(final File file) throws IOException {
    	byte[] buffer = new byte[(int)file.length()];
        InputStream input = null;
        try {
	        input = new FileInputStream(file);
	        input.read(buffer);
	        input.close();
	        return buffer;
        } catch (IOException e) {
        	if (input != null) input.close();
        	throw e;
        }
    }
    
    private String nameToPath(String name) {
    	char[] chars = name.toCharArray();
    	for (int i = 0; i < chars.length; i++) {
    		if (chars[i] == '.') chars[i] = File.separatorChar;
    	}
    	return new String(chars);
    }
    
    private String resourceToPath(String name) {
    	char[] chars = name.toCharArray();
    	for (int i = 0; i < chars.length; i++) {
    		if (chars[i] == '.') chars[i] = '/';
    	}
    	return new String(chars);
    }
    
    private static String join(String[] strings, String separator, int n) {
    	if (n <= 0) return "";
    	String res = strings[0];
    	for (int i = 1; i < n; i++) {
    		res += separator + strings[i];
    	}
    	return res;
    }
    
    private static boolean startsWith(String string, String[] prefixes) {
    	for (String prefix : prefixes) {
    		if (string.startsWith(prefix)) return true;
    	}
    	return false;
    }
}
